﻿using HeuristicLab.Algorithms.PushGP;
using HeuristicLab.Algorithms.PushGP.Expressions;
using HeuristicLab.Algorithms.PushGP.Stack;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace HeuristicLab.Tests.Interpreter.Expressions
{
    [TestClass]
    public class ExecExpressionTests : CommonTests<Expression>
    {
        protected override string TypeName { get { return "EXEC"; } }

        protected override IStack<Expression> Stack { get { return interpreter.ExecStack; } }

        protected override void Test(Expression expression)
        {
            interpreter.InterpretAsync(expression, true).Wait();
        }

        [TestMethod]
        [TestProperty("Time", "Short")]
        [TestCategory("ExpressionTest")]
        [TestCategory("ExecExpressionTest")]
        public void TestIfTrue()
        {
            interpreter.BooleanStack.Push(true);
            interpreter.Interpret("( EXEC.IF WAHR FALSCH )");

            Assert.AreEqual("WAHR", interpreter.NameStack.Top);
            TestStackCounts(nameStack: 1);
        }

        [TestMethod]
        [TestProperty("Time", "Short")]
        [TestCategory("ExpressionTest")]
        [TestCategory("ExecExpressionTest")]
        public void TestIfFalse()
        {
            interpreter.BooleanStack.Push(false);
            interpreter.Interpret("( EXEC.IF WAHR FALSCH )");

            Assert.AreEqual("FALSCH", interpreter.NameStack.Top);
            TestStackCounts(nameStack: 1);
        }

        [TestMethod]
        [TestProperty("Time", "Short")]
        [TestCategory("ExpressionTest")]
        [TestCategory("ExecExpressionTest")]
        public void TestK()
        {
            var first = Parser.Parse("A");
            var second = Parser.Parse("B");
            var third = Parser.Parse("C");

            interpreter.ExecStack.Push(third, second, first);
            Test(new ExecKExpression());

            Assert.AreEqual(first, interpreter.ExecStack.Top);
            Assert.AreEqual(third, interpreter.ExecStack.Bottom);
            TestStackCounts(execStack: 2);
        }

        [TestMethod]
        [TestProperty("Time", "Short")]
        [TestCategory("ExpressionTest")]
        [TestCategory("ExecExpressionTest")]
        public void TestS()
        {
            var first = Parser.Parse("A");
            var second = Parser.Parse("B");
            var third = Parser.Parse("C");
            var result = Parser.Parse("( B C )");

            interpreter.ExecStack.Push(third, second, first);
            Test(new ExecSExpression());

            Assert.AreEqual(result, interpreter.ExecStack.ReverseElementAt(2));
            Assert.AreEqual(third, interpreter.ExecStack.ReverseElementAt(1));
            Assert.AreEqual(first, interpreter.ExecStack.Top);
            TestStackCounts(execStack: 3);
        }

        [TestMethod]
        [TestProperty("Time", "Short")]
        [TestCategory("ExpressionTest")]
        [TestCategory("ExecExpressionTest")]
        public void TestY()
        {
            var first = Parser.Parse("A");
            var result = Parser.Parse("( EXEC.Y A )");

            interpreter.ExecStack.Push(first);
            Test(new ExecYExpression());

            Assert.AreEqual(first, interpreter.ExecStack.Top);
            Assert.AreEqual(result, interpreter.ExecStack.ReverseElementAt(1));
            TestStackCounts(execStack: 2);
        }

        [TestMethod]
        [TestProperty("Time", "Short")]
        [TestCategory("ExpressionTest")]
        [TestCategory("ExecExpressionTest")]
        public void TestNestedDoRange()
        {
            interpreter.Interpret("( 0 2 EXEC.DO*RANGE ( 1 INTEGER.+ 0 3 EXEC.DO*RANGE ( 1 INTEGER.+ INTEGER.* ) INTEGER.+ )");

            Assert.AreEqual(144, interpreter.IntegerStack.Top);
            TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        [TestProperty("Time", "Short")]
        [TestCategory("ExpressionTest")]
        [TestCategory("ExecExpressionTest")]
        public void TestNestedDoCount()
        {
            interpreter.Interpret("( 2 EXEC.DO*COUNT ( 1 INTEGER.+ 3 EXEC.DO*COUNT ( 1 INTEGER.+ INTEGER.* ) INTEGER.+ )");

            Assert.AreEqual(144, interpreter.IntegerStack.Top);
            TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        [TestProperty("Time", "Short")]
        [TestCategory("ExpressionTest")]
        [TestCategory("ExecExpressionTest")]
        public void TestNestedDoTimes()
        {
            interpreter.Interpret("( 3 EXEC.DO*TIMES ( 2 3 EXEC.DO*TIMES ( 2 INTEGER.* ) INTEGER.+ )");

            Assert.AreEqual(128, interpreter.IntegerStack.Top);
            TestStackCounts(integerStack: 1);
        }

        protected override Expression[] GetValues(int count)
        {
            var values = new Expression[count];

            for (var i = 0; i < count; i++)
            {
                values[i] = new CodeNoopExpression();
            }

            return values;
        }

        protected override Expression[] Get2Different()
        {
            return new Expression[]
            {
                new CodeNoopExpression(),
                new CodeNullExpression()
            };
        }

        protected override void CheckOtherStacksAreEmpty()
        {
            TestStackCounts(execStack: null);
        }
    }
}
